/**
 * Solution inspired by http://algo2.iti.kit.edu/heuer/kapoce/kapoce_heuristic.pdf:
 *
 * - Initial solution: N clusters, one for each node
 * - Repeat until time limit:
 *      - perturb current solution:
 *          - choose a subset of the nodes and either isolate (NODE_ISOLATOR) them or move them to a random cluster
 * (NODE_MOVER)
 *      - refine current solution:
 *          - iterate the nodes and assign each to the "cheapest cluster" (Label Propagation)
 *          - iterate the clusters and eliminate a cluster if this improves the current solutions (Clique Removal)
 *
 * - print the best solution found
 */
#include <algorithm>
#include <cassert>
#include <chrono>
#include <fstream>
#include <iostream>
#include <iterator>
#include <random>
#include <stack>
#include <unordered_map>
#include <unordered_set>
#include <vector>

#include "got2.h"

using namespace std;

#ifdef EVAL
ifstream in("input.txt");
ofstream out("output.txt");
#else
istream &in(cin);
ostream &out(cout);
#endif

#ifdef DEBUG
#define LOG(msg) clog << msg << endl
#else
#define LOG(msg)
#endif

constexpr int PRINT_TIME = 3000;
constexpr int N_REFINEMENT_LP_ITER = 10;
constexpr int N_REFINEMENT_CR_ITER = 1;
constexpr int MIN_CLIQUE_REMOVAL_SIZE = 3;
constexpr double KEEP_EQUAL_COST_SOLUTION_PROB = 0.5;
constexpr double MIN_PERTURBATION_PROB = 0.1;
constexpr double MAX_PERTURBATION_PROB = 0.25;

unsigned seed = 69420;
default_random_engine myRandomEngine(seed);
std::uniform_int_distribution<int> gen0N;
std::uniform_real_distribution<double> gen_perturbation_prob(MIN_PERTURBATION_PROB, MAX_PERTURBATION_PROB);
std::uniform_real_distribution<double> gen01(0, 1);

enum PERTURBATION_TYPE {
    NODE_ISOLATOR, NODE_MOVER
};

inline double get_perturbation_prob() { return gen_perturbation_prob(myRandomEngine); }

inline int get_rand0N() { return gen0N(myRandomEngine); }

inline double get_prob() { return gen01(myRandomEngine); }

inline long get_ms(chrono::steady_clock::time_point &begin) {
    return chrono::duration_cast<chrono::milliseconds>(chrono::steady_clock::now() - begin).count();
}

struct Cost {
    int a, r;

    bool operator<(const Cost &other) { return (*this)() < other(); }

    bool operator==(const Cost &other) { return (*this)() == other(); }

    bool operator!=(const Cost &other) { return (*this)() != other(); }

    int operator()() const { return a + r; }
};

ostream &operator<<(ostream &os, const Cost &cost) {
    os << cost.a << " " << cost.r;
    return os;
}

// number of nodes
int N;
// number of edges
int M;
// graph as adjiacency vector (for fast neighbours iteration)
vector <vector<int>> graph;
// graph as adjiacency matrix (for fast neighbour lookup)
vector <vector<bool>> graph2;

// current mapping node -> cluster_id
vector<int> cluster_id_for_node;
// current cluster_id -> nodes in that cluster
vector <unordered_set<int>> nodes_in_cluster;
// currently unused cluster ids that can be reused
stack<int> free_cluster_ids;

// data structures containing the currently best solution
Cost best_cost;
vector<int> best_cluster_id_for_node;

// shuffled to iterate the nodes in random order
vector<int> nodes_permutation;

unordered_map<int, int> count_neighbours_of_node_in_adiacent_clusters(int node_id) {
    unordered_map<int, int> adjacent_clusters_ids;

    for (int neighbour_id: graph[node_id]) {
        adjacent_clusters_ids[cluster_id_for_node[neighbour_id]]++;
    }

    return adjacent_clusters_ids;
}

int count_neighbours_of_node_in_cluster(int node_id, int cluster_id) {
    int count = 0;

    for (int neighbour_id: graph[node_id]) {
        if (cluster_id_for_node[neighbour_id] == cluster_id)
            count++;
    }

    return count;
}

Cost compute_cost_of_isolating_node(int node_id, int n_neighbours_in_same_cluster, Cost current_cost) {
    int cluster_id = cluster_id_for_node[node_id];

    int n_nodes_in_same_cluster = nodes_in_cluster[cluster_id].size();

    int n_not_neighbours_in_same_cluster =
            n_nodes_in_same_cluster - n_neighbours_in_same_cluster - 1; // subtract 1 to exclude the current node

    Cost new_cost = current_cost;
    new_cost.a -= n_not_neighbours_in_same_cluster;
    new_cost.r += n_neighbours_in_same_cluster;
    return new_cost;
    // return current_cost + n_neighbours_in_same_cluster -
    //        n_not_neighbours_in_same_cluster;
}

Cost compute_cost_of_adding_isolated_node_to_cluster(int node_id, int new_cluster_id, int n_neighbours_in_new_cluster,
                                                     Cost current_cost) {
    int cluster_id = cluster_id_for_node[node_id];
    assert(nodes_in_cluster[cluster_id].size() == 1); // assert that the node is isolated
    assert(cluster_id != new_cluster_id); // assert that the new cluster is not the one in which the node is already in

    int n_nodes_in_new_cluster = nodes_in_cluster[new_cluster_id].size();

    int n_not_neighbours_in_new_cluster = n_nodes_in_new_cluster - n_neighbours_in_new_cluster;

    Cost new_cost = current_cost;
    new_cost.a += n_not_neighbours_in_new_cluster;
    new_cost.r -= n_neighbours_in_new_cluster;
    return new_cost;
    // return current_cost - n_neighbours_in_new_cluster +
    //        n_not_neighbours_in_new_cluster;
}

void isolate_node(int node_id) {
    int cluster_id = cluster_id_for_node[node_id];

    if (nodes_in_cluster[cluster_id].size() != 1) {
        nodes_in_cluster[cluster_id].erase(node_id);

        int new_cluster_id = free_cluster_ids.top();
        free_cluster_ids.pop();

        cluster_id_for_node[node_id] = new_cluster_id;
        nodes_in_cluster[new_cluster_id].insert(node_id);
    }
}

void add_isolated_node_to_cluster(int node_id, int new_cluster_id) {
    int cluster_id = cluster_id_for_node[node_id];

    assert(nodes_in_cluster[cluster_id].size() == 1); // assert that the node is isolated
    assert(cluster_id != new_cluster_id); // assert that the new cluster is not the one in which the node is already in

    cluster_id_for_node[node_id] = new_cluster_id;
    nodes_in_cluster[new_cluster_id].insert(node_id);

    nodes_in_cluster[cluster_id].erase(node_id);
    free_cluster_ids.push(cluster_id);
}

Cost isolate_node_and_compute_cost(int node_id, int n_neighbours_in_same_cluster, Cost current_cost) {
    Cost updated_cost = compute_cost_of_isolating_node(node_id, n_neighbours_in_same_cluster, current_cost);
    isolate_node(node_id);
    return updated_cost;
}

Cost isolate_node_and_compute_cost(int node_id, Cost current_cost) {
    int n_neighbours_in_same_cluster = count_neighbours_of_node_in_cluster(node_id, cluster_id_for_node[node_id]);
    return isolate_node_and_compute_cost(node_id, n_neighbours_in_same_cluster, current_cost);
}

Cost add_isolated_node_to_cluster_and_compute_cost(int node_id, int new_cluster_id, int n_neighbours_in_new_cluster,
                                                   Cost current_cost) {
    Cost updated_cost = compute_cost_of_adding_isolated_node_to_cluster(node_id, new_cluster_id,
                                                                        n_neighbours_in_new_cluster, current_cost);
    add_isolated_node_to_cluster(node_id, new_cluster_id);
    return updated_cost;
}

Cost move_node_to_cluster_and_compute_cost(int node_id, int new_cluster_id, int n_neighbours_in_same_cluster,
                                           int n_neighbours_in_new_cluster, Cost current_cost) {
    current_cost = isolate_node_and_compute_cost(node_id, n_neighbours_in_same_cluster, current_cost);
    current_cost = add_isolated_node_to_cluster_and_compute_cost(node_id, new_cluster_id, n_neighbours_in_new_cluster,
                                                                 current_cost);
    return current_cost;
}

Cost move_node_to_random_cluster_and_compute_cost(int node_id, Cost current_cost) {
    int cluster_id = cluster_id_for_node[node_id];
    unordered_map<int, int> n_neighbours_in_adiacent_clusters = count_neighbours_of_node_in_adiacent_clusters(node_id);
    int n_neighbours_in_same_cluster = n_neighbours_in_adiacent_clusters[cluster_id];
    n_neighbours_in_adiacent_clusters.erase(cluster_id);
    if (n_neighbours_in_adiacent_clusters.size() > 0) {
        int offset = get_prob() * n_neighbours_in_adiacent_clusters.size();
        auto i_neighbour_clusters = n_neighbours_in_adiacent_clusters.begin();
        advance(i_neighbour_clusters, offset);
        int new_cluster_id = i_neighbour_clusters->first;
        int n_neighbours_in_new_cluster = n_neighbours_in_adiacent_clusters[new_cluster_id];
        current_cost = move_node_to_cluster_and_compute_cost(node_id, new_cluster_id, n_neighbours_in_same_cluster,
                                                             n_neighbours_in_new_cluster, current_cost);
    }
    return current_cost;
}

Cost compute_cluster_edits(chrono::steady_clock::time_point begin, bool print = true) {
    int a = 0, r = 0;
    if (print)
        out << best_cost.a << ' ' << best_cost.r << '\n';
    for (int node_id1 = 0; node_id1 < N; ++node_id1) {
        int cluster_id1 = best_cluster_id_for_node[node_id1];
        for (int node_id2 = node_id1 + 1; node_id2 < N; ++node_id2) {
            int cluster_id2 = best_cluster_id_for_node[node_id2];
            if (cluster_id1 == cluster_id2 && !graph2[node_id1][node_id2]) {
                if (print)
                    out << "+ " << node_id1 << ' ' << node_id2 << '\n';
                a++;
            } else if (cluster_id1 != cluster_id2 && graph2[node_id1][node_id2]) {
                if (print)
                    out << "- " << node_id1 << ' ' << node_id2 << '\n';
                r++;
            }
        }
    }
    if (print) {
        out << "***" << endl;
        out.flush();
    }
    return {a, r};
}

Cost move_to_best_target_clique(int node_id, bool consider_original_cluster, Cost current_cost) {
    int cluster_id = cluster_id_for_node[node_id];

    unordered_map<int, int> n_neighbours_in_adiacent_clusters = count_neighbours_of_node_in_adiacent_clusters(node_id);

    auto i_n_neighbours_in_same_cluster = n_neighbours_in_adiacent_clusters.find(cluster_id_for_node[node_id]);
    int n_neighbours_in_same_cluster = i_n_neighbours_in_same_cluster == n_neighbours_in_adiacent_clusters.end() ? 0
                                                                                                                 : i_n_neighbours_in_same_cluster->second;

    Cost cost_of_isolating_node = isolate_node_and_compute_cost(node_id, n_neighbours_in_same_cluster, current_cost);
    int cluster_id_of_isolated_node = cluster_id_for_node[node_id];

    Cost best_cost = cost_of_isolating_node;
    int best_new_cluster_id = cluster_id_of_isolated_node;

    for (const pair<int, int> &cluster_nodes: n_neighbours_in_adiacent_clusters) {
        int adjacent_cluster_id = cluster_nodes.first;
        int n_neighbours_in_adjacent_cluster = cluster_nodes.second;
        if (!consider_original_cluster && (adjacent_cluster_id == cluster_id)) {
            continue;
        }
        Cost cost_of_adding_node_to_adjacent_cluster = compute_cost_of_adding_isolated_node_to_cluster(node_id,
                                                                                                       adjacent_cluster_id,
                                                                                                       n_neighbours_in_adjacent_cluster,
                                                                                                       cost_of_isolating_node);

        if (cost_of_adding_node_to_adjacent_cluster < best_cost ||
            (cost_of_adding_node_to_adjacent_cluster == best_cost && get_prob() < KEEP_EQUAL_COST_SOLUTION_PROB)) {
            best_cost = cost_of_adding_node_to_adjacent_cluster;
            best_new_cluster_id = adjacent_cluster_id;
        }
    }

    if (best_new_cluster_id != cluster_id_of_isolated_node) {
        add_isolated_node_to_cluster(node_id, best_new_cluster_id);
    }

    return best_cost;
}

Cost label_propagation(Cost current_cost) {
    shuffle(nodes_permutation.begin(), nodes_permutation.end(), myRandomEngine);
    for (int node_id: nodes_permutation) {
        current_cost = move_to_best_target_clique(node_id, true, current_cost);
    }

    return current_cost;
}

Cost clique_removal(int cluster_id, Cost current_cost) {
    Cost original_cost = current_cost;
    unordered_set<int> nodes_in_original_cluster = nodes_in_cluster[cluster_id];
    for (int node_id: nodes_in_original_cluster) {
        current_cost = move_to_best_target_clique(node_id, false, current_cost);
    }
    // revert modifications if it does not improve the current solution
    if (original_cost < current_cost ||
        (current_cost == original_cost && get_prob() < 1 - KEEP_EQUAL_COST_SOLUTION_PROB)) {
        LOG("Revert clique removal");
        auto i_node_id = nodes_in_original_cluster.begin();
        isolate_node(*i_node_id);
        int new_cluster_id = cluster_id_for_node[*i_node_id];
        for (++i_node_id; i_node_id != nodes_in_original_cluster.end(); ++i_node_id) {
            isolate_node(*i_node_id);
            add_isolated_node_to_cluster(*i_node_id, new_cluster_id);
        }
        current_cost = original_cost;
    } else {
        LOG("Keep clique removal");
    }
    return current_cost;
}

Cost perturb_solution(Cost current_cost, PERTURBATION_TYPE perturbation_type) {
    double perturbation_prob = get_perturbation_prob();
    LOG("Perturbing solution of cost " << current_cost);
    shuffle(nodes_permutation.begin(), nodes_permutation.end(), myRandomEngine);
    for (int node_id: nodes_permutation) {
        if (get_prob() < perturbation_prob) {
            switch (perturbation_type) {
                case NODE_ISOLATOR:
                    current_cost = isolate_node_and_compute_cost(node_id, current_cost);
                    break;
                case NODE_MOVER:
                    current_cost = move_node_to_random_cluster_and_compute_cost(node_id, current_cost);
                    break;
            }
        }
    }
    LOG("Got solution of cost " << current_cost);
    return current_cost;
}

Cost refine_solution(Cost current_cost) {

    LOG("Refining (LP) solution of cost " << current_cost);
    /**  LABEL PROPAGATION **/
    for (int i = 0; i < N_REFINEMENT_LP_ITER; i++) {
        current_cost = label_propagation(current_cost);
    }
    LOG("Got solution of cost " << current_cost);
    LOG("Refining (CR) solution of cost " << current_cost);
    /**  CLIQUE REMOVAL **/
    for (int i = 0; i < N_REFINEMENT_CR_ITER; i++) {
        shuffle(nodes_permutation.begin(), nodes_permutation.end(), myRandomEngine);
        for (int cluster_id : nodes_permutation) {
            if (nodes_in_cluster[cluster_id].size() >= MIN_CLIQUE_REMOVAL_SIZE) {
                current_cost = clique_removal(cluster_id, current_cost);
            }
        }
    }
    LOG("Got solution of cost " << current_cost);

    return current_cost;
}

int main() {
    chrono::steady_clock::time_point begin = chrono::steady_clock::now();
    in >> N >> M;
    graph.resize(N);
    graph2.resize(N, vector<bool>(N, false));
    cluster_id_for_node.resize(N);
    nodes_in_cluster.resize(N);
    gen0N = uniform_int_distribution<int>(0, N - 1);
    nodes_permutation.resize(N);

    for (int i = 0; i < M; i++) {
        int from, to;
        in >> from >> to;
        graph[from].push_back(to);
        graph[to].push_back(from);
        graph2[from][to] = true;
        graph2[to][from] = true;
    }

    for (int i = 0; i < N; i++) {
        cluster_id_for_node[i] = i; // each node is in its own cluster
        nodes_in_cluster[i].insert(i);
        nodes_permutation[i] = i;
    }

    Cost current_cost = {0, M}; // all edges are removed

    best_cost = current_cost;
    best_cluster_id_for_node = cluster_id_for_node;
    bool improved = true;

    LOG("Started...");
    LOG("Cost: " << current_cost);
    // clog << "Computed: " << compute_cluster_edits() << endl;

    for (int i = 0;; i++) {
        // clog << "Computed: " << compute_cluster_edits() << endl;

        current_cost = refine_solution(current_cost);

        if (current_cost < best_cost) {
            best_cost = current_cost;
            best_cluster_id_for_node = cluster_id_for_node;
            improved = true;
            // int computed_cost = compute_cluster_edits();
            // assert (computed_cost == current_cost);
        }

        if (get_ms(begin) >= PRINT_TIME && improved) {
            // cerr << "printing after " << get_ms(begin) << "ms" << endl;
            Cost computed_cost = compute_cluster_edits(begin);
            LOG("Computed cost: " << computed_cost);
            assert(computed_cost == best_cost);
            improved = false;
        }
        LOG("Cost: " << current_cost);
        // clog << "Computed: " << compute_cluster_edits() << endl;
        PERTURBATION_TYPE perturbation_type =
                i % 2 == 0 ? NODE_ISOLATOR : NODE_MOVER;
        current_cost = perturb_solution(current_cost, perturbation_type);

        LOG("Cost: " << current_cost);
    }

    return 0;
}
